/* tc-i370.h -- Header file for tc-i370.c.
   Copyright 1994, 1995, 1996, 1997, 1998, 2000, 2001, 2002
   Free Software Foundation, Inc.
   Written by Ian Lance Taylor, Cygnus Support.

   This file is part of GAS, the GNU Assembler.

   GAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GAS; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.  */

#define TC_I370

#ifdef ANSI_PROTOTYPES
struct fix;
#endif

/* Set the endianness we are using.  Default to big endian.  */
#ifndef TARGET_BYTES_BIG_ENDIAN
#define TARGET_BYTES_BIG_ENDIAN 1
#endif

#ifndef BFD_ASSEMBLER
 #error I370 support requires BFD_ASSEMBLER
#endif

/* The target BFD architecture.  */
#define TARGET_ARCH (i370_arch ())
extern enum bfd_architecture i370_arch PARAMS ((void));

/* Whether or not the target is big endian */
extern int target_big_endian;

/* The target BFD format.  */
#define TARGET_FORMAT ("elf32-i370")

/* HLASM uses pseudo-ops that do not begin with a dot. */
extern bfd_boolean i370_no_pseudo_dot;
#define NO_PSEUDO_DOT i370_no_pseudo_dot

/* HLASM uses labels without a trailing colon. */
extern bfd_boolean i370_labels_without_colons;
#define LABELS_WITHOUT_COLONS i370_labels_without_colons

/* HLASM places alignment DC, DS after a label, not before it */
extern bfd_boolean i370_align_label(char *);
#define TC_START_LABEL_WITHOUT_COLON(NUL_CHAR, NEXT_CHAR) \
  (!i370_labels_without_colons || i370_align_label(line_start))

/* HLASM allows symbols starting with two @@ */
#define LEX_AT (LEX_BEGIN_NAME | LEX_NAME) /* Can have @'s inside labels.  */

/* Rewrite the @@ in HLASM symbols */
extern char * i370_canonicalize_symbol_name(char *);
#define tc_canonicalize_symbol_name i370_canonicalize_symbol_name

/* Permit temporary numeric labels.  */
#define LOCAL_LABELS_FB 1

/* Single quotes are used for literals,
 * e.g. =E'3.14159' or =XL4'cacad0d0' */
#define SINGLE_QUOTE_STRINGS

/* $ is (not) used to refer to the current location.  */
/* #define DOLLAR_DOT */

extern bfd_boolean i370_string_escapes;
#define TC_STRING_ESCAPES i370_string_escapes

/* foo-. gets turned into PC relative relocs.  */
#define DIFF_EXPR_OK

/* Values passed to md_apply_fix3 don't include the symbol value.  */
#define MD_APPLY_SYM_VALUE(FIX) 0

/* We don't need to handle .word strangely.  */
#define WORKING_DOT_WORD

/* Call md_pcrel_from_section, not md_pcrel_from.  */
#define MD_PCREL_FROM_SECTION(FIX, SEC) md_pcrel_from_section(FIX, SEC)
extern long md_pcrel_from_section PARAMS ((struct fix *, segT));

#define md_operand(x)

#define tc_comment_chars i370_comment_chars
extern const char *i370_comment_chars;

/* All 32/64-bit strings are valid float numbers. Just that some are
   denormalized. So there's no QNaN, SNaN, +inf or -inf avalable.  */
#define TC_LARGEST_EXPONENT_IS_NORMAL(PRECISION) 1

#define DWARF2_LINE_MIN_INSN_LENGTH     1
#define DWARF2_DEFAULT_RETURN_COLUMN    14  /* Link register is r14 */
#define DWARF2_CIE_DATA_ALIGNMENT       -4  /* 32-bit arch */

#define NO_STRING_ESCAPES
