/* written by Paul Edwards */
/* released to the public domain */

#ifndef UNIXIO_INCLUDED
#define UNIXIO_INCLUDED

#include <stddef.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#define S_IFBLK 0x3000
#define S_IFDIR 0x4000
#define S_IFREG 0x8000
#define S_IFMT 0xf000

#ifndef ENOENT
#define ENOENT 2002
#endif
#ifndef ENOTDIR
#define ENOTDIR 2045
#endif
#ifndef EINTR
#define EINTR 4
#endif
#ifndef EIO
#define EIO 96
#endif
#ifndef EEXIST
#define EEXIST -20 /* fix this */
#endif
#ifndef ENOTEMPTY
#define ENOTEMPTY -21 /* fix this */
#endif
#ifndef EINVAL
#define EINVAL -22 /* fix this */
#endif
#ifndef EISDIR
#define EISDIR -23 /* fix this */
#endif
#ifndef EACCES
#define EACCES -24 /* fix this */
#endif
#ifndef ENOSPC
#define ENOSPC -25 /* fix this */
#endif
#ifndef ENOMEM
#define ENOMEM -26 /* fix this */
#endif


#define O_RDONLY 1
#define O_WRONLY 2
#define O_RDWR 4
#define O_CREAT 0x100
#define O_TRUNC 0x200
#define O_EXCL 0x400
#define O_BINARY 0x800 /* fix this if wrong */

#define O_APPEND 0x1000 /* likely wrong */

struct stat {
  long st_size;
  long st_mode;
  long st_ino;
  long st_dev;
  long st_mtime;
  int st_nlink;
  int st_uid;
  int st_gid;
  int st_ctime;
  int st_atime;
};

struct timeb {
  time_t time;
  short timezone;
};

#if 0
typedef struct {
  int whatever;
} DIR;
#endif

struct dirent {
  int whatever;
};

#if 0
struct direct {
  char *d_name;
  int d_namlen;
};
#endif

typedef long off_t;

typedef long pid_t;

#define waitpid(a,b,c) (*(b) = 0)
#define getpid() (0)
#define _exit(a) (exit((a)))

#define sleep(a) (0)

int open(const char *fnm, int mode, ...);
int read(int fno, void *buf, size_t bytes);
int write(int fno, const void *buf, size_t bytes);
int close(int fno);
long lseek(int fno, long int offset, int whence);
FILE *fdopen(int fno, const char *mode);
char *mktemp(char *s);
FILE *popen(const char *s, const char *t);
int kill(int a, int b);
#if 0
DIR *opendir(const char *dirname);
/* readdir was moved and redefined in config.h to avoid
   conflict with dirent redefine in system.h */
int closedir(DIR *dir);
#endif
int execvp(const char *path, char **argv);

#endif
