# Produce executables to generate S/370 assembler


ifeq "$(targ)" "macarm"
# makefile.m64 needs to have been run in pdpclib first
# not working properly
CC=clang -S -O0 -D__ARM__ -fno-builtin -D__MACOS__ \
   -D__64BIT__ -fno-stack-protector -D__USEBIVA__ \
   -Wno-error=implicit-function-declaration
AR=ar
AS=as
LD=clang -Wl,-e, -Wl,__start
LINK2=../../pdos/pdpclib/pdpmacos.a


else ifeq "$(targ)" "gccarm64"
# makefile.a64 needs to have been run in pdpclib first
# untested
CC=gcc -S -O2 -D__ARM__ -nostdinc -nostdlib -fno-builtin \
   -D__64BIT__ -fno-stack-protector --no-pie -D__USEBIVA__
AR=ar
AS=as
LD=gcc -static -s -m64 -nostdlib
LINK1=../../pdos/pdpclib/linstart.o ../../pdos/pdpclib/pdplinux.a
LINK2=../../pdos/pdpclib/pdplinux.a


else ifeq "$(targ)" "gccamd64"
# makefile.l64 needs to have been run in pdpclib first
# I don't know why optimization needs to be off
CC=gcc -S -O0 -D__gnu_linux__ -nostdinc -nostdlib -fno-builtin \
   -D__64BIT__ -fno-stack-protector --no-pie -D__USEBIVA__ \
   -m64 -fno-common
AR=ar
AS=as
LD=gcc -s -static -m64 -nostdlib
LINK1=../../pdos/pdpclib/linstart.o ../../pdos/pdpclib/pdplinux.a
LINK2=../../pdos/pdpclib/pdplinux.a


else ifeq "$(targ)" "gccia32"
# makefile.lnx needs to have been run in pdpclib first
# testing on gcc 11.4 showed that I needed to use -O1, not -O2
# was not investigated beyond that. Could switch to gcc 3.2.3 instead
CC=gcc -S -O1 -D__gnu_linux__ -nostdinc -nostdlib -fno-builtin \
   -fno-stack-protector --no-pie -D__NOBIVA__ \
   -m32 -fno-common
AR=ar
AS=as --32
LD=gcc -s -static -m32 -nostdlib
LINK1=../../pdos/pdpclib/linstart.o ../../pdos/pdpclib/pdplinux.a
LINK2=../../pdos/pdpclib/pdplinux.a


else ifeq "$(targ)" "zpg"
# Produce z/PDOS-generic executables to generate S/370 assembler
# makefile.zpg needs to have been run in pdpclib first
# We can't detab because we can't redefine the standard
# C90 functions - but nor do we need to, because this
# is meant to be temporary assembler for use by as370
# NOFLOAT is defined because we are using an emulator that
# currently doesn't support floating point
# (NOFLOAT was deleted since mfemul now supports floating point)
CC=gccmvs -Os -U__MVS__ -D__MF32__ -D__PDOSGEN__ -I../../pdos/generic \
   -I../../pdos/src -DNO_DETAB
AS=as370 -mhlasm -mebcdic
AR=xar
LD=pdld --oformat mvs -e __crt0
LINK1=../../pdos/pdpclib/pgastart.o


else
# makefile.std needs to have been run in pdpclib first
CC=gccwin -O2 -D__WIN32__
AS=pdas --oformat coff
AR=xar
LINK1=../../pdos/pdpclib/w32start.obj
LINK2=../../pdos/pdpclib/msvcrt.lib ../../pdos/pdpclib/setjmp.obj ../../pdos/pdpclib/winsupa.obj
LD=pdld --no-insert-timestamp -s 
endif




ifeq "$(flavor)" "linux"
# This is for Linux
# Becomes gcclbf.exe
COPTS=-DTARGET_LINUX -S -fno-common -ansi -nostdinc -DHAVE_CONFIG_H \
  -DIN_GCC -DPUREISO -I ../../pdos/pdpclib -I . -DI370 \
  -D__NOBIVA__ -I config/i370 -I ../include

else ifeq "$(flavor)" "cms"
# This is for CMS
# Becomes gcccms.exe
COPTS=-DTARGET_CMS -S -fno-common -ansi -nostdinc -DHAVE_CONFIG_H \
  -DIN_GCC -DPUREISO -I ../../pdos/pdpclib -I . -DI370 \
  -D__NOBIVA__ -I config/i370 -I ../include

else ifeq "$(flavor)" "vse"
# this is for DOS/VSE
# Becomes gccvse.exe
COPTS=-DTARGET_VSE -S -fno-common -ansi -nostdinc -DHAVE_CONFIG_H \
  -DIN_GCC -DPUREISO -I ../../pdos/pdpclib -I . -DI370 \
  -D__NOBIVA__ -I config/i370 -I ../include

else ifeq "$(flavor)" "zpg"
# This is for PDOS-generic via binutils
# Becomes gcc370.exe
# This target is obsolete since we now support hlasm via gccmvs
COPTS=-DTARGET_PDOSGB -S -fno-common -ansi -nostdinc -DHAVE_CONFIG_H \
  -DIN_GCC -DPUREISO -I ../../pdos/pdpclib -I . -DI370 \
  -D__NOBIVA__ -I config/i370 -I ../include

else ifeq "$(flavor)" "zpga"
# This is for PDOS-generic via binutils as ASCII
# Becomes gcc370a.exe
COPTS=-DTARGET_ASCII -DTARGET_PDOSGB -S -fno-common -ansi -nostdinc \
  -DHAVE_CONFIG_H -DIN_GCC -DPUREISO -I ../../pdos/pdpclib -I . \
  -DI370 -D__NOBIVA__ -I config/i370 -I ../include

else
# This is for MVS
# Becomes gccmvs.exe
COPTS=-S -fno-common -ansi -nostdinc -DHAVE_CONFIG_H -DIN_GCC -DPUREISO \
  -I ../../pdos/pdpclib -I . -DI370 -D__NOBIVA__ -I config/i370 -I ../include
endif





all: clean gcc-new.exe

gcc-new.exe: \
    alias.obj \
    attribs.obj \
    bb-reorder.obj \
    bitmap.obj \
    builtins.obj \
    c-aux-info.obj \
    c-common.obj \
    c-convert.obj \
    c-decl.obj \
    c-errors.obj \
    c-format.obj \
    c-lang.obj \
    c-lex.obj \
    c-objc-common.obj \
    c-parse.obj \
    c-pragma.obj \
    c-semantics.obj \
    c-typeck.obj \
    caller-save.obj \
    calls.obj \
    cfg.obj \
    cfganal.obj \
    cfgbuild.obj \
    cfgcleanup.obj \
    cfglayout.obj \
    cfgloop.obj \
    cfgrtl.obj \
    combine.obj \
    conflict.obj \
    convert.obj \
    cppdefault.obj \
    cpperror.obj \
    cppexp.obj \
    cppfiles.obj \
    cpphash.obj \
    cppinit.obj \
    cpplex.obj \
    cpplib.obj \
    cppmacro.obj \
    cppmain.obj \
    cppspec.obj \
    cse.obj \
    cselib.obj \
    debug.obj \
    dependence.obj \
    df.obj \
    diagnostic.obj \
    dominance.obj \
    dwarf2asm.obj \
    dwarf2out.obj \
    emit-rtl.obj \
    except.obj \
    explow.obj \
    expmed.obj \
    expr.obj \
    final.obj \
    flow.obj \
    fold-const.obj \
    function.obj \
    gcc.obj \
    gccspec.obj \
    gcse.obj \
    genrtl.obj \
    ggc-common.obj \
    ggc-page.obj \
    global.obj \
    graph.obj \
    hash.obj \
    haifa-sched.obj \
    hashtable.obj \
    hooks.obj \
    ifcvt.obj \
    insn-attrtab.obj \
    insn-emit.obj \
    insn-extract.obj \
    insn-opinit.obj \
    insn-output.obj \
    insn-peep.obj \
    insn-recog.obj \
    integrate.obj \
    jump.obj \
    langhooks.obj \
    lcm.obj \
    line-map.obj \
    lists.obj \
    local-alloc.obj \
    loop.obj \
    mkdeps.obj \
    optabs.obj \
    params.obj \
    predict.obj \
    prefix.obj \
    print-rtl.obj \
    print-tree.obj \
    profile.obj \
    real.obj \
    recog.obj \
    regclass.obj \
    regmove.obj \
    regrename.obj \
    reload.obj \
    reload1.obj \
    resource.obj \
    rtl-error.obj \
    rtl.obj \
    rtlanal.obj \
    sbitmap.obj \
    sibcall.obj \
    simplify-rtx.obj \
    ssa-ccp.obj \
    ssa-dce.obj \
    ssa.obj \
    stmt.obj \
    stor-layout.obj \
    stringpool.obj \
    timevar.obj \
    toplev.obj \
    tree-dump.obj \
    tree-inline.obj \
    tree.obj \
    unroll.obj \
    varasm.obj \
    varray.obj \
    version.obj \
    ../libiberty/xmalloc.obj \
    ../libiberty/xstrerror.obj \
    ../libiberty/xstrdup.obj \
    ../libiberty/xexit.obj \
    ../libiberty/concat.obj \
    ../libiberty/hex.obj \
    ../libiberty/alloca.obj \
    ../libiberty/lbasename.obj \
    ../libiberty/obstack.obj \
    ../libiberty/strsignal.obj \
    ../libiberty/safe-ctype.obj \
    ../libiberty/splay-tree.obj \
    ../libiberty/fibheap.obj \
    ../libiberty/hashtab.obj \
    ../libiberty/asprintf.obj \
    ../libiberty/vasprintf.obj \
    ../libiberty/getpagesize.obj \
    ../libiberty/partition.obj \
    config/i370/i370.obj \
    config/i370/i370-c.obj \
    unixio.obj
  $(AR) s temp.lib
  $(LD) -o gcc-new.exe $(LINK1) temp.lib $(LINK2)

.c.obj:
  $(CC) $(COPTS) -o $*.s $<
  $(AS) -o $@ $*.s
  rm -f $*.s
  $(AR) r temp.lib $*.obj
  rm -f $*.obj

clean:
  rm -f *.s
  rm -f *.obj
  rm -f temp.lib
  rm -f gcc-new.exe




