# Produce Windows executables that use MSVCRT.DLL

# To gradually switch to 64-bit long, first change
# makefile.std in pdpclib to use pdpcrt, and change
# this makefile to link with that too. Also change
# this makefile to define LONG64. Then rebuild. Then
# modify makefile.std to use gcc-new.exe and define
# __LONG64__ and rebuild.
# Then replace gccwin.exe and pdpcrt.dll.
# Then define __LONG64__ as well as LONG64 in this
# makefile and rebuild. Then replace gccwin.exe.
# Then rebuild. Then replace gccwin.exe and rebuild
# and the executables should be identical (fc /b).
# Then rebuild pdpcrt.dll and confirm it is identical.
# Or maybe replace and repeat?

# By default, this is set up to reproduce gccwin itself,
# a Win32 executable


ifeq "$(targ)" "zpg"
CC=gccmvs
else
CC=gccwin
#CC=gccwin_l64
#CC=gccw64_l64
#CC=gccw64
endif


ifeq "$(targ)" "zpg"
AS=as370 -mhlasm -mebcdic
else
# If using gccw64, use the --64 version
AS=pdas --oformat coff
#AS=pdas --oformat coff --64
endif


AR=xar


ifeq "$(targ)" "zpg"
LD=pdld --oformat mvs -e __crt0
else
LD=pdld
endif


ifeq "$(targ)" "zpg"
EXTRA=../../pdos/pdpclib/zpgsupa.o
else
# If using gccw64, use x64supb.obj
EXTRA=../../pdos/pdpclib/winsupa.obj ../../pdos/pdpclib/setjmp.obj
#EXTRA=../../pdos/pdpclib/x64supb.obj
endif


CFLAGS=

# define this if you want the compiler to PRODUCE 64-bit long
#CFLAGS:=$(CFLAGS) -DLONG64

# define this once you ALREADY have 64-bit long
#CFLAGS:=$(CFLAGS) -D__LONG64__

# define this if you want the compiler to PRODUCE 64-bit int
#CFLAGS:=$(CFLAGS) -DINT64

# define this once you ALREADY have 64-bit int
#CFLAGS:=$(CFLAGS) -D__INT64__

# define this if you want the compiler to PRODUCE 32-bit short
#CFLAGS:=$(CFLAGS) -DSHORT32

# define this once you ALREADY have 32-bit short
#CFLAGS:=$(CFLAGS) -D__SHORT32__

# define this if you ALREADY have 64-bit pointers
# (ie the compiler is already producing 64-bit pointers, even
# if it is still a 32-bit executable)
#CFLAGS:=$(CFLAGS) -D__64BIT__

# define this if you are on an EBCDIC system and
# want to target an EBCDIC system
ifeq "$(tgtchs)" "ebcdic"
CFLAGS:=$(CFLAGS) -DTARGET_EBCDIC -DHOST_EBCDIC -funsigned-char
#
# Or use this if you are on an ASCII system wanting to target EBCDIC
CFLAGS:=$(CFLAGS) -DTARGET_EBCDIC
endif

ifeq "$(targ)" "zpg"
# This is if you want to produce a z/PDOS-generic executable
# Must have run makefile.zpg in PDPCLIB first
# Note that gccmvs as a cross-compiler from ASCII to EBCDIC
# isn't good enough to cope with all the charset implications,
# so you will need to compile this on the mainframe instead.
COPTS=$(CFLAGS) -Os -U__MVS__ -D__MF32__ -D__PDOSGEN__ -I../../pdos/generic \
   -I../../pdos/src -DNO_DETAB \
   -S -D__NOBIVA__ -DI386 -DHAVE_CONFIG_H -DIN_GCC -DPUREISO \
   -I ../../pdos/pdpclib -I . -I config/i386 -I config -I ../include \
   -DUSE_SHORT_NAMES -I config/i370 \
   -DUSE_C_ALLOCA -DHOST_EBCDIC -DTARGET_EBCDIC
else
# This is to build gccwin - with optional 64-bit long
COPTS=$(CFLAGS) -fno-common -O2 -S -D__WIN32__ -D__NOBIVA__ -DI386 -DHAVE_CONFIG_H -DIN_GCC -DPUREISO -I ../../pdos/pdpclib -I ../../pdos/src -I . -I config/i386 -I config -I ../include

# This is if you want to build gccw64_l64 that uses Windows calling convention
# Or gccw64 (implied _l32) - choose using the LONG64 define above
#COPTS=$(CFLAGS) -DTARGET_64BIT_DEFAULT -DMSABI -fno-common -O2 -S -D__WIN32__ -D__NOBIVA__ -DI386 -DHAVE_CONFIG_H -DIN_GCC -DPUREISO -I ../../pdos/pdpclib -I ../../pdos/src -I . -I config/i386 -I config -I ../include

# Ditto, but using gccw64_l64 itself
# You will need to add -D__STATIC__ and makefile.p64 and change the link step
# Not sure if the -DVALIST_NOT_PTR is correct
# It seems that the simplistic vararg routines in PDPCLIB are not
# good enough to handle the complex gcc code on x64. And also the
# gcc builtins are not good enough to handle 32-bit long. So you
# can't use gccw64 (and PDPCLIB makefile.s64) for this currently
#COPTS=$(CFLAGS) -DVALIST_NOT_PTR -D__USEBIVA__ -fno-builtin -DTARGET_64BIT_DEFAULT -DMSABI -fno-common -O2 -S -D__WIN32__ -DI386 -DHAVE_CONFIG_H -DIN_GCC -DPUREISO -I ../../pdos/pdpclib -I ../../pdos/src -I . -I config/i386 -I config -I ../include

# This is if you want to build gcc386
#COPTS=$(CFLAGS) -fno-common -O2 -S -D__WIN32__ -D__NOBIVA__ -DI386 -DBUILDAOUT -DHAVE_CONFIG_H -DIN_GCC -DPUREISO -I ../../pdos/pdpclib -I ../../pdos/src -I . -I config/i386 -I config -I ../include
endif



all: clean gcc-new.exe

gcc-new.exe: \
    alias.o \
    attribs.o \
    bb-reorder.o \
    bitmap.o \
    builtins.o \
    c-aux-info.o \
    c-common.o \
    c-convert.o \
    c-decl.o \
    c-errors.o \
    c-format.o \
    c-lang.o \
    c-lex.o \
    c-objc-common.o \
    c-parse.o \
    c-pragma.o \
    c-semantics.o \
    c-typeck.o \
    caller-save.o \
    calls.o \
    cfg.o \
    cfganal.o \
    cfgbuild.o \
    cfgcleanup.o \
    cfglayout.o \
    cfgloop.o \
    cfgrtl.o \
    combine.o \
    conflict.o \
    convert.o \
    cppdefault.o \
    cpperror.o \
    cppexp.o \
    cppfiles.o \
    cpphash.o \
    cppinit.o \
    cpplex.o \
    cpplib.o \
    cppmacro.o \
    cppmain.o \
    cppspec.o \
    cse.o \
    cselib.o \
    debug.o \
    dependence.o \
    df.o \
    diagnostic.o \
    dominance.o \
    dwarf2asm.o \
    dwarf2out.o \
    emit-rtl.o \
    except.o \
    explow.o \
    expmed.o \
    expr.o \
    final.o \
    flow.o \
    fold-const.o \
    function.o \
    gcc.o \
    gccspec.o \
    gcse.o \
    genrtl.o \
    ggc-common.o \
    ggc-page.o \
    global.o \
    graph.o \
    hash.o \
    haifa-sched.o \
    hashtable.o \
    hooks.o \
    ifcvt.o \
    insn-attrtab.o \
    insn-emit.o \
    insn-extract.o \
    insn-opinit.o \
    insn-output.o \
    insn-peep.o \
    insn-recog.o \
    integrate.o \
    jump.o \
    langhooks.o \
    lcm.o \
    line-map.o \
    lists.o \
    local-alloc.o \
    loop.o \
    mkdeps.o \
    optabs.o \
    params.o \
    predict.o \
    prefix.o \
    print-rtl.o \
    print-tree.o \
    profile.o \
    real.o \
    recog.o \
    regclass.o \
    regmove.o \
    regrename.o \
    reload.o \
    reload1.o \
    resource.o \
    rtl-error.o \
    rtl.o \
    rtlanal.o \
    sbitmap.o \
    sched-deps.o \
    sched-rgn.o \
    sched-vis.o \
    sibcall.o \
    simplify-rtx.o \
    ssa-ccp.o \
    ssa-dce.o \
    ssa.o \
    stmt.o \
    stor-layout.o \
    stringpool.o \
    timevar.o \
    toplev.o \
    tree-dump.o \
    tree-inline.o \
    tree.o \
    unroll.o \
    varasm.o \
    varray.o \
    version.o \
    ../libiberty/xmalloc.o \
    ../libiberty/xstrerror.o \
    ../libiberty/xstrdup.o \
    ../libiberty/xexit.o \
    ../libiberty/concat.o \
    ../libiberty/hex.o \
    ../libiberty/alloca.o \
    ../libiberty/lbasename.o \
    ../libiberty/obstack.o \
    ../libiberty/strsignal.o \
    ../libiberty/safe-ctype.o \
    ../libiberty/splay-tree.o \
    ../libiberty/fibheap.o \
    ../libiberty/hashtab.o \
    ../libiberty/asprintf.o \
    ../libiberty/vasprintf.o \
    ../libiberty/getpagesize.o \
    ../libiberty/partition.o \
    config/i386/i386.o \
    config/i386/winnt.o \
    unixio.o \
    reg-stack.o \
    doloop.o \
    sdbout.o \
    dbxout.o \
    ../libiberty/md5.o
  $(AR) s temp.a
ifeq "$(targ)" "zpg"
# for z/pdos-generic target:
  $(LD) -s -o gcc-new.exe ../../pdos/pdpclib/pgastart.o temp.a $(EXTRA)
else
# A stack size of 0xf000 is insufficient to compile c-common.c in gcc 3.4.6 (3.2.3 is fine)
# (ie attempting to compile gcc 3.4.6 source code using the gcc 3.2.3 compiler)
# 3f000 is also insufficient
# 4f000 is sufficient
# but setting to 5f000 for good measure
  pdld --no-insert-timestamp --stack 0,0x5f000 -s -o gcc-new.exe ../../pdos/pdpclib/w32start.obj temp.a ../../pdos/pdpclib/msvcrt.lib $(EXTRA)
#  xar x temp.a gcc.o
#  pdld --no-insert-timestamp --stack 0,0x5f000 -s -o gcc-new.exe ../../pdos/pdpclib/p32start.obj gcc.o temp.a ../../pdos/pdpclib/pdpwin64.lib ../../pdos/src/kernel32.lib
endif

.c.o:
  $(CC) $(COPTS) -o $*.s $<
  $(AS) -o $@ $*.s
  rm -f $*.s
  $(AR) r temp.a $*.o
  rm -f $*.o

clean:
  rm -f *.s
  rm -f *.o
  rm -f temp.a
  rm -f gcc-new.exe
