# Produce Windows executable using Visual Studio

# Note that if we enable optimization with -Ox or -O2,
# the executable that is produced
# is getting this internal error for some files:
# stdio.c:2862: Internal compiler error in mark_jump_label, at jump.c:1489

# So we leave it disabled, ie -Od

# Problem was traced to optimization of just reload1.c
# Then further traced to just function reload_cse_simplify_operands
# Putting #pragma optimize("g", off)
# before this function and then putting
# #pragma optimize("g", on)
# after, is enough to work around the problem.
# But the root cause appears to be the unconditional
# use of alloca() here:
# op_alt_regno[i] = (int *) alloca (recog_data.n_alternatives
# which likely places constraints on the optimizer's ability to
# optimize, and alloca probably releases memory that it
# should not be.

# Note that this optimization issue only seems to apply to
# 80836 not ARM, so we have ARM set to do optimization

# Note that there are warnings that likely no-one is interested in fixing,
# so we specify -w to suppress them

# Build with pdmake -B -f windowsm.mak
# nmake can also be used

# You will need to have built rm.exe from OZPD first

# You can target either Win32 for 80386 or Win32 for ARM

# I believe you also need to choose whether you are using
# the Visual C runtime, or PDPCLIB (using makefile.msx I think)

CC=cl
# -Ox

# for 80386
#COPTS=-Od -w -nologo -Za -c -D__NOBIVA__ -DI386 -DHAVE_CONFIG_H -DIN_GCC -DPUREISO -I . -I config/i386 -I config -I ../include
#EXTRA=config/i386/i386.obj config/i386/winnt.obj
#LIB=lib -machine:x86

# for ARM32
#COPTS=-Od -w -nologo -Za -c -D__NOBIVA__ -DBUILDPRM -DHAVE_CONFIG_H -DHAVE__BOOL -DIN_GCC -DPUREISO -I . -I config/arm -I config -I ../include
#EXTRA=config/arm/arm.obj config/arm/pe.obj
#LIB=lib -machine:arm

# for ARM32 using pdld and pdpclib
# Also need to modify the link
# Note that you need to run makefile.mar first
COPTS=-Gs2147483647 -D__MSC__ -D__WIN32__ -GS- -Zl -D__ARM__ -I../../pdos/pdpclib -O2 -w -nologo -Za -c -D__NOBIVA__ -DBUILDPRM -DHAVE_CONFIG_H -DHAVE__BOOL -DIN_GCC -DPUREISO -I . -I config/arm -I config -I ../include
EXTRA=config/arm/arm.obj config/arm/pe.obj
LIB=lib -machine:arm
#LD=link
LD=pdld --no-insert-timestamp

all: clean gcc-new.exe

gcc-new.exe: \
    alias.obj \
    attribs.obj \
    bb-reorder.obj \
    bitmap.obj \
    builtins.obj \
    c-aux-info.obj \
    c-common.obj \
    c-convert.obj \
    c-decl.obj \
    c-errors.obj \
    c-format.obj \
    c-lang.obj \
    c-lex.obj \
    c-objc-common.obj \
    c-parse.obj \
    c-pragma.obj \
    c-semantics.obj \
    c-typeck.obj \
    caller-save.obj \
    calls.obj \
    cfg.obj \
    cfganal.obj \
    cfgbuild.obj \
    cfgcleanup.obj \
    cfglayout.obj \
    cfgloop.obj \
    cfgrtl.obj \
    combine.obj \
    conflict.obj \
    convert.obj \
    cppdefault.obj \
    cpperror.obj \
    cppexp.obj \
    cppfiles.obj \
    cpphash.obj \
    cppinit.obj \
    cpplex.obj \
    cpplib.obj \
    cppmacro.obj \
    cppmain.obj \
    cppspec.obj \
    cse.obj \
    cselib.obj \
    debug.obj \
    dependence.obj \
    df.obj \
    diagnostic.obj \
    dominance.obj \
    dwarf2asm.obj \
    dwarf2out.obj \
    emit-rtl.obj \
    except.obj \
    explow.obj \
    expmed.obj \
    expr.obj \
    final.obj \
    flow.obj \
    fold-const.obj \
    function.obj \
    gcc.obj \
    gccspec.obj \
    gcse.obj \
    genrtl.obj \
    ggc-common.obj \
    ggc-page.obj \
    global.obj \
    graph.obj \
    hash.obj \
    haifa-sched.obj \
    hashtable.obj \
    hooks.obj \
    ifcvt.obj \
    insn-attrtab.obj \
    insn-emit.obj \
    insn-extract.obj \
    insn-opinit.obj \
    insn-output.obj \
    insn-peep.obj \
    insn-recog.obj \
    integrate.obj \
    jump.obj \
    langhooks.obj \
    lcm.obj \
    line-map.obj \
    lists.obj \
    local-alloc.obj \
    loop.obj \
    mkdeps.obj \
    optabs.obj \
    params.obj \
    predict.obj \
    prefix.obj \
    print-rtl.obj \
    print-tree.obj \
    profile.obj \
    real.obj \
    recog.obj \
    regclass.obj \
    regmove.obj \
    regrename.obj \
    reload.obj \
    reload1.obj \
    resource.obj \
    rtl-error.obj \
    rtl.obj \
    rtlanal.obj \
    sbitmap.obj \
    sched-deps.obj \
    sched-rgn.obj \
    sched-vis.obj \
    sibcall.obj \
    simplify-rtx.obj \
    ssa-ccp.obj \
    ssa-dce.obj \
    ssa.obj \
    stmt.obj \
    stor-layout.obj \
    stringpool.obj \
    timevar.obj \
    toplev.obj \
    tree-dump.obj \
    tree-inline.obj \
    tree.obj \
    unroll.obj \
    varasm.obj \
    varray.obj \
    version.obj \
    ../libiberty/xmalloc.obj \
    ../libiberty/xstrerror.obj \
    ../libiberty/xstrdup.obj \
    ../libiberty/xexit.obj \
    ../libiberty/concat.obj \
    ../libiberty/hex.obj \
    ../libiberty/alloca.obj \
    ../libiberty/lbasename.obj \
    ../libiberty/obstack.obj \
    ../libiberty/strsignal.obj \
    ../libiberty/safe-ctype.obj \
    ../libiberty/splay-tree.obj \
    ../libiberty/fibheap.obj \
    ../libiberty/hashtab.obj \
    ../libiberty/asprintf.obj \
    ../libiberty/vasprintf.obj \
    ../libiberty/getpagesize.obj \
    ../libiberty/partition.obj \
    $(EXTRA) \
    unixio.obj \
    reg-stack.obj \
    doloop.obj \
    sdbout.obj \
    dbxout.obj \
    ../libiberty/md5.obj
####    $(CC) -nologo -Fegcc-new.exe gcc.obj temp.lib -link -fixed:no
    $(LD) --stack 0,0x5f000 -s -nostdlib -o gcc-new.exe -e __pdpent ../../pdos/pdpclib/p32start.obj gcc.obj ../../pdos/pdpclib/armsupb.obj ../../pdos/pdpclib/armsupc.obj temp.lib ../../pdos/pdpclib/pdpwin32.lib ../../pdos/src/kernel32.lib
#    $(LD) -nologo -fixed:no -out:gcc-new.exe ../../pdos/pdpclib/w32start.obj gcc.obj ../../pdos/pdpclib/armsupb.obj ../../pdos/pdpclib/armsupc.obj temp.lib ../../pdos/pdpclib/msvcrt.lib
####    $(LD) --stack 0,0x5f000 -s -nostdlib -o gcc-new.exe -e __pdpent ../../pdos/pdpclib/w32start.obj gcc.obj ../../pdos/pdpclib/armsupb.obj ../../pdos/pdpclib/armsupc.obj temp.lib ../../pdos/pdpclib/msvcrt.lib

.c.obj:
  $(CC) $(COPTS) -Fo$@ $<
  $(LIB) -nologo -out:temp.lib temp.lib $@

clean:
  rm -f *.obj
  rm -f temp.lib
  copy nul: temp.def
  $(LIB) -nologo -out:temp.lib -def:temp.def
  rm -f temp.def
  rm -f gcc-new.exe
