/*
  Copyright (c) 1990-2005 Info-ZIP.  All rights reserved.

  See the accompanying file LICENSE, version 2000-Apr-09 or later
  (the contents of which are also included in unzip.h) for terms of use.
  If, for some reason, all these files are missing, the Info-ZIP license
  also may be found at:  ftp://ftp.info-zip.org/pub/infozip/license.html
*/
/*---------------------------------------------------------------------------

  pdos.c (for PDOS/86 and PDOS/386)

  Contains:
             open_outfile()
             close_outfile()
             close_infile()
             do_wild()
             mapattr()
             mapname()
             checkdir()
             check_for_newer()
             stat()
             version()

  ---------------------------------------------------------------------------*/


#define __PDOS_C       /* identifies this source module */
#define UNZIP_INTERNAL
#include "unzip.h"

#ifdef __WIN32__
#include "windows.h"

#define ISDIR    FILE_ATTRIBUTE_DIRECTORY
#endif

#ifdef __PDOS386__
#include "pos.h"

#define ISDIR    FILE_ATTR_DIRECTORY
#endif

#define S_IFMT  0xF000
#define S_IFREG 0x8000
#define S_IFDIR 0x4000

#ifdef S_ISDIR
#undef S_ISDIR
#define S_ISDIR(m)  (((m)& S_IFMT) == S_IFDIR)
#endif


#ifndef ENOENT
#define ENOENT 2002
#endif

static int created_dir;        /* used by mapname(), checkdir() */
static int renamed_fullpath;   /* ditto */
static unsigned nLabelDrive;   /* ditto, plus volumelabel() */

/****************************/
/* Function close_outfile() */
/****************************/

void close_outfile(__G)
__GDEF
{
    fclose(G.outfile);
} /* end function close_outfile() */


/***************************/
/* Function close_infile() */
/***************************/

void close_infile(__G)
__GDEF
{
    fclose(G.zipfd);
} /* end function close_infile() */



/************************/
/*  Function mapattr()  */
/************************/

int mapattr(__G)
__GDEF
{
    return 0;
}


/************************/
/*  Function mapname()  */
/************************/

int mapname(__G__ renamed)
__GDEF
int renamed;
/*
 * returns:
 *  MPN_OK          - no problem detected
 *  MPN_INF_TRUNC   - caution (truncated filename)
 *  MPN_INF_SKIP    - info "skip entry" (dir doesn't exist)
 *  MPN_ERR_SKIP    - error -> skip entry
 *  MPN_ERR_TOOLONG - error -> path is too long
 *  MPN_NOMEM       - error (memory allocation failed) -> skip entry
 *  [also MPN_VOL_LABEL, MPN_CREATED_DIR]
 */
{
    char pathcomp[FILNAMSIZ];      /* path-component buffer */
    char *pp, *cp = (char *)NULL;    /* character pointers */
    char *lastsemi = (char *)NULL;   /* pointer to last semi-colon in pathcomp */
    int killed_ddot = FALSE;       /* is set when skipping "../" pathcomp */
    int error = MPN_OK;
    register unsigned workch;      /* hold the character being tested */


/*---------------------------------------------------------------------------
    Initialize various pointers and counters and stuff.
  ---------------------------------------------------------------------------*/

  /* can create path as long as not just freshening, or if user told us */
    G.create_dirs = (!uO.fflag || renamed);

    created_dir = FALSE;        /* not yet */
    renamed_fullpath = FALSE;

    if (renamed) {
        cp = G.filename - 1;    /* point to beginning of renamed name... */
        while (*++cp)
            if (*cp == '\\')    /* convert backslashes to forward */
                *cp = '/';
        cp = G.filename;
        /* use temporary rootpath if user gave full pathname */
        if (G.filename[0] == '/') {
            renamed_fullpath = TRUE;
            pathcomp[0] = '/';  /* copy the '/' and terminate */
            pathcomp[1] = '\0';
            ++cp;
        }
        else if (isalpha((uch)G.filename[0]) && G.filename[1] == ':') {
            renamed_fullpath = TRUE;
            pp = pathcomp;
            *pp++ = *cp++;      /* copy the "d:" (+ '/', possibly) */
            *pp++ = *cp++;
            if (*cp == '/')
                *pp++ = *cp++;  /* otherwise add "./"? */
            *pp = '\0';
        }
    }

    /* pathcomp is ignored unless renamed_fullpath is TRUE: */
    if ((error = checkdir(__G__ pathcomp, INIT)) != 0) /* initialize path buf */
        return error;           /* ...unless no mem or vol label on hard disk */

    *pathcomp = '\0';           /* initialize translation buffer */
    pp = pathcomp;              /* point to translation buffer */
    if (!renamed) {             /* cp already set if renamed */
        if (uO.jflag)           /* junking directories */
            cp = (char *)strrchr(G.filename, '/');
        if (cp == (char *)NULL) /* no '/' or not junking dirs */
            cp = G.filename;    /* point to internal zipfile-member pathname */
        else
            ++cp;               /* point to start of last component of path */
    }

    /*---------------------------------------------------------------------------
        Begin main loop through characters in filename.
      ---------------------------------------------------------------------------*/

    while ((workch = (uch)*cp++) != 0) {

        switch (workch) {
        case '/':             /* can assume -j flag not given */
            *pp = '\0';
            if (strcmp(pathcomp, ".") == 0) {
                /* don't bother appending "./" to the path */
                *pathcomp = '\0';
            }
            else if (!uO.ddotflag && strcmp(pathcomp, "..") == 0) {
                /* "../" dir traversal detected, skip over it */
                *pathcomp = '\0';
                killed_ddot = TRUE;     /* set "show message" flag */
            }
            /* when path component is not empty, append it now */
            if (*pathcomp != '\0' &&
                ((error = checkdir(__G__ pathcomp, APPEND_DIR))
                    & MPN_MASK) > MPN_INF_TRUNC)
                return error;
            pp = pathcomp;    /* reset conversion buffer for next piece */
            lastsemi = (char *)NULL; /* leave direct. semi-colons alone */
            break;

            /* drive names are not stored in zipfile, so no colons allowed;
             *  no brackets or most other punctuation either (all of which
             *  can appear in Unix-created archives; backslash is particularly
             *  bad unless all necessary directories exist) */

        case ':':           /* special shell characters of command.com */
        case '\\':          /*  (device and directory limiters, wildcard */
        case '"':           /*  characters, stdin/stdout redirection and */
        case '<':           /*  pipe indicators and the quote sign) are */
        case '>':           /*  never allowed in filenames on (V)FAT */
        case '|':
        case '*':
        case '?':
            *pp++ = '_';
            break;

        case ';':             /* start of VMS version? */
            lastsemi = pp;
            break;

        default:
            /* allow ASCII 255 and European characters in filenames: */
            if (isprint(workch) || workch >= 127)
                *pp++ = (char)workch;

        } /* end switch */
    } /* end while loop */

    /*---------------------------------------------------------------------------
        Report if directory was created (and no file to create:  filename ended
        in '/'), check name to be sure it exists, and combine path and name be-
        fore exiting.
      ---------------------------------------------------------------------------*/

    if (G.filename[strlen(G.filename) - 1] == '/')
    {
        checkdir(__G__ G.filename, GETPATH);
        if (created_dir) {
            if (QCOND2) {
                Info(slide, 0, ((char *)slide, "   creating: %s\n", FnFilter1(G.filename)));
            }
            return (error & ~MPN_MASK) | MPN_CREATED_DIR;
        }
        else if (IS_OVERWRT_ALL) {
            /* overwrite attributes of existing directory on user's request */

            /* set file attributes: */
            /*z_dos_chmod(__G__ G.filename, G.pInfo->file_attr);*/
        }
        /* dir existed already; don't look for data to extract */
        return (error & ~MPN_MASK) | MPN_INF_SKIP;
    }

    *pp = '\0';                   /* done with pathcomp:  terminate it */

    /* if not saving them, remove VMS version numbers (appended ";###") */
    if (!uO.V_flag && lastsemi) {
        while (isdigit((uch)(*pp)))
            ++pp;
        if (*pp == '\0')          /* only digits between ';' and end:  nuke */
            *lastsemi = '\0';
    }

    if (G.pInfo->vollabel) {
        if (strlen(pathcomp) > 11)
            pathcomp[11] = '\0';
    }

    if (*pathcomp == '\0') {
        Info(slide, 1, ((char *)slide, "mapname:  conversion of %s failed\n",
            FnFilter1(G.filename)));
        return (error & ~MPN_MASK) | MPN_ERR_SKIP;
    }

    checkdir(__G__ pathcomp, APPEND_NAME);  /* returns 1 if truncated: care? */
    checkdir(__G__ G.filename, GETPATH);

    if (G.pInfo->vollabel) {    /* set the volume label now */
        if (QCOND2)
            Info(slide, 0, ((char *)slide, "labelling %c: %-22s\n",
            (nLabelDrive + 'a' - 1),
                FnFilter1(G.filename)));
        if (mkdir(G.filename)) {
            return (error & ~MPN_MASK) | MPN_ERR_SKIP;
        }
        /* success:  skip the "extraction" quietly */
        return (error & ~MPN_MASK) | MPN_INF_SKIP;
    }
    return error;
} /* end function mapname() */



/******************************/
/* Function check_for_newer() */  /* used for overwriting/freshening/updating */
/******************************/

int check_for_newer(__G__ filename)  /* return 1 if existing file is newer */
__GDEF                           /*  or equal; 0 if older; -1 if doesn't */
char *filename;                  /*  exist yet */
{
    FILE *stream;

    if ((stream = fopen(filename, FOPR)) != NULL) {
        fclose(stream);
        /* File exists, assume it is "newer" than archive entry. */
        return EXISTS_AND_NEWER;
    }
    /* File does not exist. */
    return DOES_NOT_EXIST;
} /* end function check_for_newer() */



int checkdir(__G__ pathcomp, flag)
__GDEF
char *pathcomp;
int flag;
/*
 * returns:
 *  MPN_OK          - no problem detected
 *  MPN_INF_TRUNC   - (on APPEND_NAME) truncated filename
 *  MPN_INF_SKIP    - path doesn't exist, not allowed to create
 *  MPN_ERR_SKIP    - path doesn't exist, tried to create and failed; or path
 *                    exists and is not a directory, but is supposed to be
 *  MPN_ERR_TOOLONG - path is too long
 *  MPN_NOMEM       - can't allocate memory for filename buffers
 */
{
    static int rootlen = 0;   /* length of rootpath */
    static char *rootpath;    /* user's "extract-to" directory */
    static char *buildpath;   /* full path (so far) to extracted file */
    static char *end;         /* pointer to end of buildpath ('\0') */

#   define FN_MASK   7
#   define FUNCTION  (flag & FN_MASK)

/*---------------------------------------------------------------------------
    APPEND_DIR:  append the path component to the path being built and check
    for its existence.  If doesn't exist and we are creating directories, do
    so for this one; else signal success or error as appropriate.
  ---------------------------------------------------------------------------*/

    if (FUNCTION == APPEND_DIR) {
        int too_long = FALSE;
        Trace((stderr, "appending dir segment [%s]\n", FnFilter1(pathcomp)));
        while ((*end = *pathcomp++) != '\0')
            ++end;

        /* GRR:  could do better check, see if overrunning buffer as we go:
         * check end-buildpath after each append, set warning variable if
         * within 20 of FILNAMSIZ; then if var set, do careful check when
         * appending.  Clear variable when begin new path. */

        if ((end - buildpath) > FILNAMSIZ - 3)  /* need '/', one-char name, '\0' */
            too_long = TRUE;                /* check if extracting directory? */
        if (SSTAT(buildpath, &G.statbuf))   /* path doesn't exist */
        {
            if (!G.create_dirs) { /* told not to create (freshening) */
                free(buildpath);
                /* path doesn't exist:  nothing to do */
                return MPN_INF_SKIP;
            }
            if (too_long) {
                free(buildpath);
                /* no room for filenames:  fatal */
                return MPN_ERR_TOOLONG;
            }
            if (mkdir(buildpath, 0777) == -1) {   /* create the directory */
                free(buildpath);
                /* path didn't exist, tried to create, failed */
                return MPN_ERR_SKIP;
            }
            created_dir = TRUE;
        }
        else if (!S_ISDIR(G.statbuf.st_mode)) {
            free(buildpath);
            /* path existed but wasn't dir */
            return MPN_ERR_SKIP;
        }
        if (too_long) {
            free(buildpath);
            /* no room for filenames:  fatal */
            return MPN_ERR_TOOLONG;
        }
        *end++ = '/';
        *end = '\0';
        Trace((stderr, "buildpath now = [%s]\n", FnFilter1(buildpath)));
        return MPN_OK;

    } /* end if (FUNCTION == APPEND_DIR) */

/*---------------------------------------------------------------------------
    GETPATH:  copy full path to the string pointed at by pathcomp, and free
    buildpath.
  ---------------------------------------------------------------------------*/

    if (FUNCTION == GETPATH) {
        strcpy(pathcomp, buildpath);
        Trace((stderr, "getting and freeing path [%s]\n",
            FnFilter1(pathcomp)));
        free(buildpath);
        buildpath = end = (char *)NULL;
        return MPN_OK;
    }

    /*---------------------------------------------------------------------------
        APPEND_NAME:  assume the path component is the filename; append it and
        return without checking for existence.
      ---------------------------------------------------------------------------*/

    if (FUNCTION == APPEND_NAME) {
#ifdef NOVELL_BUG_WORKAROUND
        if (end == buildpath && !G.pInfo->vollabel) {
            /* work-around for Novell's "overwriting executables" bug:
               prepend "./" to name when no path component is specified */
            *end++ = '.';
            *end++ = '/';
        }
#endif /* NOVELL_BUG_WORKAROUND */
        Trace((stderr, "appending filename [%s]\n", FnFilter1(pathcomp)));
        while ((*end = *pathcomp++) != '\0') {
            ++end;
            if ((end - buildpath) >= FILNAMSIZ) {
                *--end = '\0';
                return MPN_INF_TRUNC;   /* filename truncated */
            }
        }
        Trace((stderr, "buildpath now = [%s]\n", FnFilter1(buildpath)));
        /* could check for existence here, prompt for new name... */
        return MPN_OK;
    }

    /*---------------------------------------------------------------------------
        INIT:  allocate and initialize buffer space for the file currently being
        extracted.  If file was renamed with an absolute path, don't prepend the
        extract-to path.
      ---------------------------------------------------------------------------*/

    if (FUNCTION == INIT) {
        Trace((stderr, "initializing buildpath to "));
        /* allocate space for full filename, root path, and maybe "./" */
        if ((buildpath = (char *)malloc(strlen(G.filename) + rootlen + 3)) ==
            (char *)NULL)
            return MPN_NOMEM;
        if (G.pInfo->vollabel) {
            /* GRR:  for network drives, do strchr() and return IZ_VOL_LABEL if not [1] */
            if (renamed_fullpath && pathcomp[1] == ':')
                *buildpath = (char)ToLower(*pathcomp);
            else if (!renamed_fullpath && rootlen > 1 && rootpath[1] == ':')
                *buildpath = (char)ToLower(*rootpath);
            else {
                /*GETDRIVE(nLabelDrive);   /* assumed that a == 1, b == 2, etc. */
                *buildpath = (char)(nLabelDrive - 1 + 'a');
            }
            nLabelDrive = *buildpath - 'a' + 1;        /* save for mapname() */
            *buildpath = '\0';
            end = buildpath;
        }
        else if (renamed_fullpath) {   /* pathcomp = valid data */
            end = buildpath;
            while ((*end = *pathcomp++) != '\0')
                ++end;
        }
        else if (rootlen > 0) {
            strcpy(buildpath, rootpath);
            end = buildpath + rootlen;
        }
        else {
            *buildpath = '\0';
            end = buildpath;
        }
        Trace((stderr, "[%s]\n", FnFilter1(buildpath)));
        return MPN_OK;
    }

    /*---------------------------------------------------------------------------
        ROOT:  if appropriate, store the path in rootpath and create it if neces-
        sary; else assume it's a zipfile member and return.  This path segment
        gets used in extracting all members from every zipfile specified on the
        command line.  Note that under OS/2 and MS-DOS, if a candidate extract-to
        directory specification includes a drive letter (leading "x:"), it is
        treated just as if it had a trailing '/'--that is, one directory level
        will be created if the path doesn't exist, unless this is otherwise pro-
        hibited (e.g., freshening).
      ---------------------------------------------------------------------------*/

#if (!defined(SFX) || defined(SFX_EXDIR))
    if (FUNCTION == ROOT) {
        Trace((stderr, "initializing root path to [%s]\n",
            FnFilter1(pathcomp)));
        if (pathcomp == (char *)NULL) {
            rootlen = 0;
            return MPN_OK;
        }
        if (rootlen > 0)        /* rootpath was already set, nothing to do */
            return MPN_OK;
        if ((rootlen = strlen(pathcomp)) > 0) {
            int had_trailing_pathsep = FALSE, has_drive = FALSE, add_dot = FALSE;
            char *tmproot;
            if ((tmproot = (char *)malloc(rootlen + 3)) == (char *)NULL) {
                rootlen = 0;
                return MPN_NOMEM;
            }
            strcpy(tmproot, pathcomp);
            if (isalpha((uch)tmproot[0]) && tmproot[1] == ':')
                has_drive = TRUE;   /* drive designator */
            if (tmproot[rootlen - 1] == '/' || tmproot[rootlen - 1] == '\\') {
                tmproot[--rootlen] = '\0';
                had_trailing_pathsep = TRUE;
            }
            if (has_drive && (rootlen == 2)) {
                if (!had_trailing_pathsep)   /* i.e., original wasn't "x:/" */
                    add_dot = TRUE;    /* relative path: add '.' before '/' */
            }
            else if (rootlen > 0) {     /* need not check "x:." and "x:/" */
                if (SSTAT(tmproot, &G.statbuf) || !S_ISDIR(G.statbuf.st_mode))
                {
                    /* path does not exist */
                    if (!G.create_dirs /* || iswild(tmproot) */) {
                        free(tmproot);
                        rootlen = 0;
                        /* treat as stored file */
                        return MPN_INF_SKIP;
                    }
                    /* GRR:  scan for wildcard characters?  OS-dependent...  if find any, return 2:
                     * treat as stored file(s) */
                     /* create directory (could add loop here scanning tmproot
                      * to create more than one level, but really necessary?) */
                    if (mkdir(tmproot, 0777) == -1) {
                        free(tmproot);
                        rootlen = 0;
                        /* path didn't exist, tried to create, failed: */
                        /* file exists, or need 2+ subdir levels */
                        return MPN_ERR_SKIP;
                    }
                }
            }
            if (add_dot)                    /* had just "x:", make "x:." */
                tmproot[rootlen++] = '.';
            tmproot[rootlen++] = '/';
            tmproot[rootlen] = '\0';
            if ((rootpath = (char *)realloc(tmproot, rootlen + 1)) == NULL) {
                free(tmproot);
                rootlen = 0;
                return MPN_NOMEM;
            }
            Trace((stderr, "rootpath now = [%s]\n", FnFilter1(rootpath)));
        }
        return MPN_OK;
    }
#endif /* !SFX || SFX_EXDIR */

    /*---------------------------------------------------------------------------
        END:  free rootpath, immediately prior to program exit.
      ---------------------------------------------------------------------------*/
    if (FUNCTION == END) {
        Trace((stderr, "freeing rootpath\n"));
        if (rootlen > 0) {
            free(rootpath);
            rootlen = 0;
        }
        return MPN_OK;
    }
    return MPN_INVALID; /* should never reach */
} /* end function checkdir() */



int stat(char *f, struct stat *buf)
{
    FILE *fp;
    int attr;
    memset(buf, '\0', sizeof *buf);
#ifdef __WIN32__
    attr = GetFileAttributes(f);
#endif
#ifdef __PDOS386__
    PosGetFileAttributes(f, &attr);
#endif
    if (attr == ISDIR)
    {
        buf->st_mode = S_IFDIR;
        return (0);
    }
    else
    {
        if (attr < 0)
        {
            return (-1);
        }
        buf->st_mode = S_IFREG;
        fp = fopen(f, "rb");
        if (fp == NULL)
        {
            errno = ENOENT;
            return (-1);
        }
        fseek(fp, 0, SEEK_END);
        buf->st_size = ftell(fp);
        fclose(fp);
        return (0);
    }
}

/**********************/
/* Function  mkdir()  */
/**********************/

int mkdir(const char *dname)
{
#ifdef __WIN32__
    return (CreateDirectory(dname, NULL));
#endif
#ifdef __PDOS386__
    return (PosMakeDir(dname));
#endif
}

/**********************/
/* Function do_wild() */   /* for porting: dir separator; match(ignore_case) */
/**********************/

char *do_wild(__G__ wld)
__GDEF
ZCONST char *wld;      /* only used first time on a given dir */
{
    static int First = 0;
    static char filename[256];

    if (First == 0) {
        First = 1;
        strncpy(filename, wld, sizeof(filename));
        filename[sizeof(filename) - 1] = '\0';
        return filename;
    }
    else
        return (char *)NULL;

} /* end function do_wild() */



/************************/
/*  Function version()  */
/************************/

void version(__G)
__GDEF
{
    printf("need to implement version()\n");
} /* end function version() */
